/*

Copyright (c) 2018-2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include "ustring.h"
#include <stdlib.h>

// Unicode replacement character for invalid code points
#define REPL_8_1  '\xEF'  // UTF-8, first code unit
#define REPL_8_2  '\xBF'  // UTF-8, second code unit
#define REPL_8_3  '\xBD'  // UTF-8, third code unit
#define REPL_CPT  0xFFFDU // code point for UTF-16 and UTF-32


typedef struct tag_CPCHAR // element type of code page arrays
{
  uint32_t codepoint;
  int      charcode;
} CPCHAR;

int cp1252order = 0; // ascending sort order of the Windows-1252 code page array; 0 = by charcode, 1 = by codepoint
CPCHAR cp1252[256] =   // code page Windows-1252
{
  /*      |      _0      |       _1      |       _2      |       _3      |       _4      |       _5      |       _6      |       _7      |       _8      |       _9      |       _A      |       _B      |       _C      |       _D      |       _E      |       _F      */
  /* 0_ */ {0x0000U,   0}, {0x0001U,   1}, {0x0002U,   2}, {0x0003U,   3}, {0x0004U,   4}, {0x0005U,   5}, {0x0006U,   6}, {0x0007U,   7}, {0x0008U,   8}, {0x0009U,   9}, {0x000AU,  10}, {0x000BU,  11}, {0x000CU,  12}, {0x000DU,  13}, {0x000EU,  14}, {0x000FU,  15},
  /* 1_ */ {0x0010U,  16}, {0x0011U,  17}, {0x0012U,  18}, {0x0013U,  19}, {0x0014U,  20}, {0x0015U,  21}, {0x0016U,  22}, {0x0017U,  23}, {0x0018U,  24}, {0x0019U,  25}, {0x001AU,  26}, {0x001BU,  27}, {0x001CU,  28}, {0x001DU,  29}, {0x001EU,  30}, {0x001FU,  31},
  /* 2_ */ {0x0020U,  32}, {0x0021U,  33}, {0x0022U,  34}, {0x0023U,  35}, {0x0024U,  36}, {0x0025U,  37}, {0x0026U,  38}, {0x0027U,  39}, {0x0028U,  40}, {0x0029U,  41}, {0x002AU,  42}, {0x002BU,  43}, {0x002CU,  44}, {0x002DU,  45}, {0x002EU,  46}, {0x002FU,  47},
  /* 3_ */ {0x0030U,  48}, {0x0031U,  49}, {0x0032U,  50}, {0x0033U,  51}, {0x0034U,  52}, {0x0035U,  53}, {0x0036U,  54}, {0x0037U,  55}, {0x0038U,  56}, {0x0039U,  57}, {0x003AU,  58}, {0x003BU,  59}, {0x003CU,  60}, {0x003DU,  61}, {0x003EU,  62}, {0x003FU,  63},
  /* 4_ */ {0x0040U,  64}, {0x0041U,  65}, {0x0042U,  66}, {0x0043U,  67}, {0x0044U,  68}, {0x0045U,  69}, {0x0046U,  70}, {0x0047U,  71}, {0x0048U,  72}, {0x0049U,  73}, {0x004AU,  74}, {0x004BU,  75}, {0x004CU,  76}, {0x004DU,  77}, {0x004EU,  78}, {0x004FU,  79},
  /* 5_ */ {0x0050U,  80}, {0x0051U,  81}, {0x0052U,  82}, {0x0053U,  83}, {0x0054U,  84}, {0x0055U,  85}, {0x0056U,  86}, {0x0057U,  87}, {0x0058U,  88}, {0x0059U,  89}, {0x005AU,  90}, {0x005BU,  91}, {0x005CU,  92}, {0x005DU,  93}, {0x005EU,  94}, {0x005FU,  95},
  /* 6_ */ {0x0060U,  96}, {0x0061U,  97}, {0x0062U,  98}, {0x0063U,  99}, {0x0064U, 100}, {0x0065U, 101}, {0x0066U, 102}, {0x0067U, 103}, {0x0068U, 104}, {0x0069U, 105}, {0x006AU, 106}, {0x006BU, 107}, {0x006CU, 108}, {0x006DU, 109}, {0x006EU, 110}, {0x006FU, 111},
  /* 7_ */ {0x0070U, 112}, {0x0071U, 113}, {0x0072U, 114}, {0x0073U, 115}, {0x0074U, 116}, {0x0075U, 117}, {0x0076U, 118}, {0x0077U, 119}, {0x0078U, 120}, {0x0079U, 121}, {0x007AU, 122}, {0x007BU, 123}, {0x007CU, 124}, {0x007DU, 125}, {0x007EU, 126}, {0x007FU, 127},
  /* 8_ */ {0x20ACU, 128}, {0x0081U, 129}, {0x201AU, 130}, {0x0192U, 131}, {0x201EU, 132}, {0x2026U, 133}, {0x2020U, 134}, {0x2021U, 135}, {0x02C6U, 136}, {0x2030U, 137}, {0x0160U, 138}, {0x2039U, 139}, {0x0152U, 140}, {0x008DU, 141}, {0x017DU, 142}, {0x008FU, 143},
  /* 9_ */ {0x0090U, 144}, {0x2018U, 145}, {0x2019U, 146}, {0x201CU, 147}, {0x201DU, 148}, {0x2022U, 149}, {0x2013U, 150}, {0x2014U, 151}, {0x02DCU, 152}, {0x2122U, 153}, {0x0161U, 154}, {0x203AU, 155}, {0x0153U, 156}, {0x009DU, 157}, {0x017EU, 158}, {0x0178U, 159},
  /* A_ */ {0x00A0U, 160}, {0x00A1U, 161}, {0x00A2U, 162}, {0x00A3U, 163}, {0x00A4U, 164}, {0x00A5U, 165}, {0x00A6U, 166}, {0x00A7U, 167}, {0x00A8U, 168}, {0x00A9U, 169}, {0x00AAU, 170}, {0x00ABU, 171}, {0x00ACU, 172}, {0x00ADU, 173}, {0x00AEU, 174}, {0x00AFU, 175},
  /* B_ */ {0x00B0U, 176}, {0x00B1U, 177}, {0x00B2U, 178}, {0x00B3U, 179}, {0x00B4U, 180}, {0x00B5U, 181}, {0x00B6U, 182}, {0x00B7U, 183}, {0x00B8U, 184}, {0x00B9U, 185}, {0x00BAU, 186}, {0x00BBU, 187}, {0x00BCU, 188}, {0x00BDU, 189}, {0x00BEU, 190}, {0x00BFU, 191},
  /* C_ */ {0x00C0U, 192}, {0x00C1U, 193}, {0x00C2U, 194}, {0x00C3U, 195}, {0x00C4U, 196}, {0x00C5U, 197}, {0x00C6U, 198}, {0x00C7U, 199}, {0x00C8U, 200}, {0x00C9U, 201}, {0x00CAU, 202}, {0x00CBU, 203}, {0x00CCU, 204}, {0x00CDU, 205}, {0x00CEU, 206}, {0x00CFU, 207},
  /* D_ */ {0x00D0U, 208}, {0x00D1U, 209}, {0x00D2U, 210}, {0x00D3U, 211}, {0x00D4U, 212}, {0x00D5U, 213}, {0x00D6U, 214}, {0x00D7U, 215}, {0x00D8U, 216}, {0x00D9U, 217}, {0x00DAU, 218}, {0x00DBU, 219}, {0x00DCU, 220}, {0x00DDU, 221}, {0x00DEU, 222}, {0x00DFU, 223},
  /* E_ */ {0x00E0U, 224}, {0x00E1U, 225}, {0x00E2U, 226}, {0x00E3U, 227}, {0x00E4U, 228}, {0x00E5U, 229}, {0x00E6U, 230}, {0x00E7U, 231}, {0x00E8U, 232}, {0x00E9U, 233}, {0x00EAU, 234}, {0x00EBU, 235}, {0x00ECU, 236}, {0x00EDU, 237}, {0x00EEU, 238}, {0x00EFU, 239},
  /* F_ */ {0x00F0U, 240}, {0x00F1U, 241}, {0x00F2U, 242}, {0x00F3U, 243}, {0x00F4U, 244}, {0x00F5U, 245}, {0x00F6U, 246}, {0x00F7U, 247}, {0x00F8U, 248}, {0x00F9U, 249}, {0x00FAU, 250}, {0x00FBU, 251}, {0x00FCU, 252}, {0x00FDU, 253}, {0x00FEU, 254}, {0x00FFU, 255}
};

int cp850order = 0; // ascending sort order of the CP 850 code page array; 0 = by charcode, 1 = by codepoint
CPCHAR cp850[256] =   // code page 850
{
  /*      |      _0      |       _1      |       _2      |       _3      |       _4      |       _5      |       _6      |       _7      |       _8      |       _9      |       _A      |       _B      |       _C      |       _D      |       _E      |       _F      */
  /* 0_ */ {0x0000U,   0}, {0x0001U,   1}, {0x0002U,   2}, {0x0003U,   3}, {0x0004U,   4}, {0x0005U,   5}, {0x0006U,   6}, {0x0007U,   7}, {0x0008U,   8}, {0x0009U,   9}, {0x000AU,  10}, {0x000BU,  11}, {0x000CU,  12}, {0x000DU,  13}, {0x000EU,  14}, {0x000FU,  15},
  /* 1_ */ {0x0010U,  16}, {0x0011U,  17}, {0x0012U,  18}, {0x0013U,  19}, {0x0014U,  20}, {0x0015U,  21}, {0x0016U,  22}, {0x0017U,  23}, {0x0018U,  24}, {0x0019U,  25}, {0x001AU,  26}, {0x001BU,  27}, {0x001CU,  28}, {0x001DU,  29}, {0x001EU,  30}, {0x001FU,  31},
  /* 2_ */ {0x0020U,  32}, {0x0021U,  33}, {0x0022U,  34}, {0x0023U,  35}, {0x0024U,  36}, {0x0025U,  37}, {0x0026U,  38}, {0x0027U,  39}, {0x0028U,  40}, {0x0029U,  41}, {0x002AU,  42}, {0x002BU,  43}, {0x002CU,  44}, {0x002DU,  45}, {0x002EU,  46}, {0x002FU,  47},
  /* 3_ */ {0x0030U,  48}, {0x0031U,  49}, {0x0032U,  50}, {0x0033U,  51}, {0x0034U,  52}, {0x0035U,  53}, {0x0036U,  54}, {0x0037U,  55}, {0x0038U,  56}, {0x0039U,  57}, {0x003AU,  58}, {0x003BU,  59}, {0x003CU,  60}, {0x003DU,  61}, {0x003EU,  62}, {0x003FU,  63},
  /* 4_ */ {0x0040U,  64}, {0x0041U,  65}, {0x0042U,  66}, {0x0043U,  67}, {0x0044U,  68}, {0x0045U,  69}, {0x0046U,  70}, {0x0047U,  71}, {0x0048U,  72}, {0x0049U,  73}, {0x004AU,  74}, {0x004BU,  75}, {0x004CU,  76}, {0x004DU,  77}, {0x004EU,  78}, {0x004FU,  79},
  /* 5_ */ {0x0050U,  80}, {0x0051U,  81}, {0x0052U,  82}, {0x0053U,  83}, {0x0054U,  84}, {0x0055U,  85}, {0x0056U,  86}, {0x0057U,  87}, {0x0058U,  88}, {0x0059U,  89}, {0x005AU,  90}, {0x005BU,  91}, {0x005CU,  92}, {0x005DU,  93}, {0x005EU,  94}, {0x005FU,  95},
  /* 6_ */ {0x0060U,  96}, {0x0061U,  97}, {0x0062U,  98}, {0x0063U,  99}, {0x0064U, 100}, {0x0065U, 101}, {0x0066U, 102}, {0x0067U, 103}, {0x0068U, 104}, {0x0069U, 105}, {0x006AU, 106}, {0x006BU, 107}, {0x006CU, 108}, {0x006DU, 109}, {0x006EU, 110}, {0x006FU, 111},
  /* 7_ */ {0x0070U, 112}, {0x0071U, 113}, {0x0072U, 114}, {0x0073U, 115}, {0x0074U, 116}, {0x0075U, 117}, {0x0076U, 118}, {0x0077U, 119}, {0x0078U, 120}, {0x0079U, 121}, {0x007AU, 122}, {0x007BU, 123}, {0x007CU, 124}, {0x007DU, 125}, {0x007EU, 126}, {0x007FU, 127},
  /* 8_ */ {0x00C7U, 128}, {0x00FCU, 129}, {0x00E9U, 130}, {0x00E2U, 131}, {0x00E4U, 132}, {0x00E0U, 133}, {0x00E5U, 134}, {0x00E7U, 135}, {0x00EAU, 136}, {0x00EBU, 137}, {0x00E8U, 138}, {0x00EFU, 139}, {0x00EEU, 140}, {0x00ECU, 141}, {0x00C4U, 142}, {0x00C5U, 143},
  /* 9_ */ {0x00C9U, 144}, {0x00E6U, 145}, {0x00C6U, 146}, {0x00F4U, 147}, {0x00F6U, 148}, {0x00F2U, 149}, {0x00FBU, 150}, {0x00F9U, 151}, {0x00FFU, 152}, {0x00D6U, 153}, {0x00DCU, 154}, {0x00F8U, 155}, {0x00A3U, 156}, {0x00D8U, 157}, {0x00D7U, 158}, {0x0192U, 159},
  /* A_ */ {0x00E1U, 160}, {0x00EDU, 161}, {0x00F3U, 162}, {0x00FAU, 163}, {0x00F1U, 164}, {0x00D1U, 165}, {0x00AAU, 166}, {0x00BAU, 167}, {0x00BFU, 168}, {0x00AEU, 169}, {0x00ACU, 170}, {0x00BDU, 171}, {0x00BCU, 172}, {0x00A1U, 173}, {0x00ABU, 174}, {0x00BBU, 175},
  /* B_ */ {0x2591U, 176}, {0x2592U, 177}, {0x2593U, 178}, {0x2502U, 179}, {0x2524U, 180}, {0x00C1U, 181}, {0x00C2U, 182}, {0x00C0U, 183}, {0x00A9U, 184}, {0x2563U, 185}, {0x2551U, 186}, {0x2557U, 187}, {0x255DU, 188}, {0x00A2U, 189}, {0x00A5U, 190}, {0x2510U, 191},
  /* C_ */ {0x2514U, 192}, {0x2534U, 193}, {0x252CU, 194}, {0x251CU, 195}, {0x2500U, 196}, {0x253CU, 197}, {0x00E3U, 198}, {0x00C3U, 199}, {0x255AU, 200}, {0x2554U, 201}, {0x2569U, 202}, {0x2566U, 203}, {0x2560U, 204}, {0x2550U, 205}, {0x256CU, 206}, {0x00A4U, 207},
  /* D_ */ {0x00F0U, 208}, {0x00D0U, 209}, {0x00CAU, 210}, {0x00CBU, 211}, {0x00C8U, 212}, {0x0131U, 213}, {0x00CDU, 214}, {0x00CEU, 215}, {0x00CFU, 216}, {0x2518U, 217}, {0x250CU, 218}, {0x2588U, 219}, {0x2584U, 220}, {0x00A6U, 221}, {0x00CCU, 222}, {0x2580U, 223},
  /* E_ */ {0x00D3U, 224}, {0x00DFU, 225}, {0x00D4U, 226}, {0x00D2U, 227}, {0x00F5U, 228}, {0x00D5U, 229}, {0x00B5U, 230}, {0x00FEU, 231}, {0x00DEU, 232}, {0x00DAU, 233}, {0x00DBU, 234}, {0x00D9U, 235}, {0x00FDU, 236}, {0x00DDU, 237}, {0x00AFU, 238}, {0x00B4U, 239},
  /* F_ */ {0x00ADU, 240}, {0x00B1U, 241}, {0x2017U, 242}, {0x00BEU, 243}, {0x00B6U, 244}, {0x00A7U, 245}, {0x00F7U, 246}, {0x00B8U, 247}, {0x00B0U, 248}, {0x00A8U, 249}, {0x00B7U, 250}, {0x00B9U, 251}, {0x00B3U, 252}, {0x00B2U, 253}, {0x25A0U, 254}, {0x00A0U, 255}
};

int CharCodeCmp(const void *a, const void *b) // charcode comparison; callback function to be used for qsort
{
  if (((const CPCHAR*)a)->charcode > ((const CPCHAR*)b)->charcode) return 1;
  if (((const CPCHAR*)a)->charcode < ((const CPCHAR*)b)->charcode) return -1;
  return 0;
}

int CodePointCmp(const void *a, const void *b) // codepoint comparison; callback function to be used for qsort and bsearch
{
  if (((const CPCHAR*)a)->codepoint > ((const CPCHAR*)b)->codepoint) return 1;
  if (((const CPCHAR*)a)->codepoint < ((const CPCHAR*)b)->codepoint) return -1;
  return 0;
}

uint16_t *ReverseByteOrder16(uint16_t *const str)
{
  for (uint16_t *it = str; *it != 0U; ++it)
    *it = (*it >> 8U) | (*it << 8U);

  return str;
}

uint32_t *ReverseByteOrder32(uint32_t *const str)
{
  for (uint32_t *it = str; *it != 0U; ++it)
    *it = (*it >> 24U) | ((*it & 0x00FF0000U) >> 8U) | ((*it & 0x0000FF00U) << 8U) | (*it << 24U);

  return str;
}

int HasBOM8(const char *const str)
{
  if (*str == BOM_8_1 && str[1] == BOM_8_2 && str[2] == BOM_8_3)
    return 1;

  return 0;
}

int HasBOM16(const uint16_t *const str)
{
  if (*str == BOM_CPT)
    return 1;

  if (*str == 0xFFFEU)
    return -1;

  return 0;
}

int HasBOM32(const uint32_t *const str)
{
  if (*str == BOM_CPT)
    return 1;

  if (*str == 0xFFFE0000U)
    return -1;

  return 0;
}

size_t Validate8(const char *const str, size_t *const p_len)
{
  const uint8_t *it = (const uint8_t *const)str;
  size_t length = 0U, bytes = 0U;

  for (; *it != 0U; ++it, ++length, ++bytes)
  {
    // 01..7F is ASCII and always valid
    if (*it > 0x7FU)
    {
      // valid two bytes
      // - C2..DF | 80..BF (first byte 0xC0 and 0xC1 invalid)
      if (*it >= 0xC2U && *it <= 0xDFU &&
          it[1] >= 0x80U && it[1] <= 0xBFU)
      {
        ++it;
        ++bytes;
      }
      // valid three bytes
      // - E0     | A0..BF | 80..BF
      // - E1..EC | 80..BF | 80..BF
      // - ED     | 80..9F | 80..BF
      // - EE..EF | 80..BF | 80..BF
      else if (((
                  *it == 0xE0U &&
                  it[1] >= 0xA0U && it[1] <= 0xBFU) ||
                (
                  *it >= 0xE1U && *it <= 0xECU &&
                  it[1] >= 0x80U && it[1] <= 0xBFU) ||
                (
                  *it == 0xEDU &&
                  it[1] >= 0x80U && it[1] <= 0x9FU) ||
                (
                  *it >= 0xEEU && *it <= 0xEFU &&
                  it[1] >= 0x80U && it[1] <= 0xBFU)) &&
               it[2] >= 0x80U && it[2] <= 0xBFU)
      {
        it += 2;
        bytes += 2U;
      }
      // valid four bytes
      // - F0     | 90..BF | 80..BF | 80..BF
      // - F1..F3 | 80..BF | 80..BF | 80..BF
      // - F4     | 80..8F | 80..BF | 80..BF
      else if (((
                  *it == 0xF0U &&
                  it[1] >= 0x90U && it[1] <= 0xBFU) ||
                (
                  *it >= 0xF1U && *it <= 0xF3U &&
                  it[1] >= 0x80U && it[1] <= 0xBFU) ||
                (
                  *it == 0xF4U &&
                  it[1] >= 0x80U && it[1] <= 0x8FU)) &&
               it[2] >= 0x80U && it[2] <= 0xBFU &&
               it[3] >= 0x80U && it[3] <= 0xBFU)
      {
        it += 3;
        bytes += 3U;
      }
      else
        return 0U;
    }
  }

  if (p_len != NULL)
    *p_len = length;

  return bytes + 1U;
}

size_t Validate16(const uint16_t *const str, size_t *const p_len)
{
  const uint16_t *it = str;
  size_t length = 0U, bytes = 0U;

  for (; *it != 0U; ++it, ++length, bytes += 2U)
  {
    if (*it > 0xD7FFU && *it < 0xDC00U)
    {
      ++it;
      bytes += 2U;
      if (*it < 0xDC00U || *it > 0xDFFFU)
        return 0U;
    }
  }

  if (p_len != NULL)
    *p_len = length;

  return bytes + 2U;
}

size_t Validate32(const uint32_t *const str, size_t *const p_len)
{
  size_t length = 0U, bytes = 0U;
  const uint32_t *it = str;

  for (; *it != 0U; ++it, ++length, bytes += 4U)
    if ((*it > 0x0000D7FFU && *it < 0x0000E000U) || *it > 0x0010FFFFU)
      return 0U;

  if (p_len != NULL)
    *p_len = length;

  return bytes + 4U;
}

size_t Cp850ToUtf32(const char *const str, uint32_t *const buf, const size_t bufsiz)
{
  if (cp850order != 0)
  {
    qsort(&cp850[128], 128U, sizeof(CPCHAR), CharCodeCmp);
    cp850order = 0;
  }

  size_t bytes = 4U;
  const uint8_t *it_in = (const uint8_t *const)str;
  uint32_t *it_out = buf;

  while (*it_in != 0U)
  {
    if (bufsiz < (bytes += 4U) && buf != NULL)
      return 0U;

    if (buf != NULL)
      *it_out++ = cp850[*it_in].codepoint;

    ++it_in;
  }

  *it_out = 0U;
  return bytes;
}

size_t Cp1252ToUtf32(const char *const str, uint32_t *const buf, const size_t bufsiz)
{
  if (cp1252order != 0)
  {
    qsort(&cp1252[128], 128U, sizeof(CPCHAR), CharCodeCmp);
    cp1252order = 0;
  }

  size_t bytes = 4U;
  const uint8_t *it_in = (const uint8_t *const)str;
  uint32_t *it_out = buf;

  while (*it_in != 0U)
  {
    if (bufsiz < (bytes += 4U) && buf != NULL)
      return 0U;

    if (buf != NULL)
      *it_out++ = cp1252[*it_in].codepoint;

    ++it_in;
  }

  *it_out = 0U;
  return bytes;
}

size_t Utf8To32(const char *const str, uint32_t *const buf, const size_t bufsiz)
{
  size_t bytes = 4U;
  const uint8_t *it_in = (const uint8_t *const)str;
  uint32_t *it_out = buf;

  while (*it_in != 0U)
  {
    if (bufsiz < (bytes += 4U) && buf != NULL)
      return 0U;

    if (*it_in < 0x80U)
    {
      if (buf != NULL)
        *it_out++ = *it_in;

      if (*it_in != 0U)
        ++it_in;
    }
    else if (*it_in >= 0xC2U && *it_in <= 0xDFU)
    {
      if (buf != NULL)
      {
        if (it_in[1] >= 0x80U && it_in[1] <= 0xBFU)
          *it_out++ = ((*it_in ^ 0x000000C0U) << 6U) | (it_in[1] ^ 0x00000080U);
        else
          *it_out++ = REPL_CPT;
      }

      ++it_in;
      if (*it_in != 0U)
        ++it_in;
    }
    else if (*it_in >= 0xE0U && *it_in <= 0xEFU)
    {
      if (buf != NULL)
      {
        if (((
               *it_in == 0xE0U &&
               it_in[1] >= 0xA0U && it_in[1] <= 0xBFU) ||
             (
               *it_in >= 0xE1U && *it_in <= 0xECU &&
               it_in[1] >= 0x80U && it_in[1] <= 0xBFU) ||
             (
               *it_in == 0xEDU &&
               it_in[1] >= 0x80U && it_in[1] <= 0x9FU) ||
             (
               *it_in >= 0xEEU && *it_in <= 0xEFU &&
               it_in[1] >= 0x80U && it_in[1] <= 0xBFU)) &&
            it_in[2] >= 0x80U && it_in[2] <= 0xBFU)
        {
          *it_out = ((*it_in ^ 0x000000E0U) << 12U) | ((it_in[1] ^ 0x00000080U) << 6U) | (it_in[2] ^ 0x00000080U);
          ++it_out;
        }
        else
          *it_out++ = REPL_CPT;
      }

      ++it_in;
      for (unsigned i = 0U; i < 2U && *it_in != 0U; ++it_in, ++i);
    }
    else if (*it_in >= 0xF0U && *it_in <= 0xF4U)
    {
      if (buf != NULL)
      {
        if (((
               *it_in == 0xF0U &&
               it_in[1] >= 0x90U && it_in[1] <= 0xBFU) ||
             (
               *it_in >= 0xF1U && *it_in <= 0xF3U &&
               it_in[1] >= 0x80U && it_in[1] <= 0xBFU) ||
             (
               *it_in == 0xF4U &&
               it_in[1] >= 0x80U && it_in[1] <= 0x8FU)) &&
            it_in[2] >= 0x80U && it_in[2] <= 0xBFU &&
            it_in[3] >= 0x80U && it_in[3] <= 0xBFU)
        {
          *it_out = ((*it_in ^ 0x000000F0U) << 18U) | ((it_in[1] ^ 0x00000080U) << 12U) | ((it_in[2] ^ 0x00000080U) << 6U) | (it_in[3] ^ 0x00000080U);
          ++it_out;
        }
        else
          *it_out++ = REPL_CPT;
      }

      ++it_in;
      for (unsigned i = 0U; i < 3U && *it_in != 0U; ++it_in, ++i);
    }
    else
    {
      if (buf != NULL)
        *it_out++ = REPL_CPT;

      ++it_in;
    }
  }

  *it_out = 0U;
  return bytes;
}

size_t Utf16To32(const uint16_t *const str, uint32_t *const buf, const size_t bufsiz)
{
  size_t bytes = 4U;
  const uint16_t *it_in = str;
  uint32_t *it_out = buf;

  while (*it_in != 0U)
  {
    if (bufsiz < (bytes += 4U) && buf != NULL)
      return 0U;

    if (*it_in > 0xD7FFU && *it_in < 0xDC00U)
    {
      const uint32_t high = *it_in++;
      if (buf != NULL)
      {
        if (*it_in > 0xDBFFU && *it_in < 0xE000U)
          *it_out++ = (high << 10U) + *it_in - 0x035FDC00U;
        else
          *it_out++ = REPL_CPT;
      }
    }
    else if (buf != NULL)
      *it_out++ = *it_in;

    if (*it_in != 0U)
      ++it_in;
  }

  *it_out = 0U;
  return bytes;
}

size_t Utf32ToCP850(const uint32_t *const str, char *const buf, const size_t bufsiz)
{
  if (cp850order == 0)
  {
    qsort(&cp850[128], 128U, sizeof(CPCHAR), CodePointCmp);
    cp850order = 1;
  }

  size_t bytes = 1U;
  const uint32_t *it_in = str;
  uint8_t *it_out = (uint8_t *const)buf;

  while (*it_in != 0U)
  {
    if (bufsiz < ++bytes && buf != NULL)
      return 0U;

    if (buf != NULL)
    {
      if (*it_in == REPL_CPT)
        *it_out++ = 63;
      else if (*it_in < 128U)
        *it_out++ = (uint8_t) * it_in;
      else
      {
        CPCHAR *found = (CPCHAR*)bsearch(it_in, &cp850[128], 128U, sizeof(CPCHAR), CodePointCmp);
        *it_out++ = found == NULL ? 63 : found->charcode;
      }
    }

    ++it_in;
  }

  *it_out = 0U;
  return bytes;
}

size_t Utf32ToCP1252(const uint32_t *const str, char *const buf, const size_t bufsiz)
{
  if (cp1252order == 0)
  {
    qsort(&cp1252[128], 128U, sizeof(CPCHAR), CodePointCmp);
    cp1252order = 1;
  }

  size_t bytes = 1U;
  const uint32_t *it_in = str;
  uint8_t *it_out = (uint8_t *const)buf;

  while (*it_in != 0U)
  {
    if (bufsiz < ++bytes && buf != NULL)
      return 0U;

    if (buf != NULL)
    {
      if (*it_in == REPL_CPT)
        *it_out++ = 63;
      else if (*it_in < 128U)
        *it_out++ = (uint8_t) * it_in;
      else
      {
        CPCHAR *found = (CPCHAR*)bsearch(it_in, &cp1252[128], 128U, sizeof(CPCHAR), CodePointCmp);
        *it_out++ = found == NULL ? 63 : found->charcode;
      }
    }

    ++it_in;
  }

  *it_out = 0U;
  return bytes;
}

size_t Utf32To8(const uint32_t *const str, char *const buf, const size_t bufsiz)
{
  size_t bytes = 1U;
  const uint32_t *it_in = str;
  uint8_t *it_out = (uint8_t *const)buf;

  while (*it_in != 0U)
  {
    if (*it_in < 0x00000080U)
    {
      if (bufsiz < ++bytes && buf != NULL)
        return 0U;

      if (buf != NULL)
        *it_out++ = *it_in;

      ++it_in;
    }
    else if (*it_in < 0x00000800U)
    {
      if (bufsiz < (bytes += 2U) && buf != NULL)
        return 0U;

      if (buf != NULL)
      {
        *it_out++ = (*it_in >> 6U & 0x1FU) | 0xC0U;
        *it_out++ = (*it_in & 0x3FU) | 0x80U;
      }

      ++it_in;
    }
    else if ((*it_in > 0x0000D7FFU && *it_in < 0x0000E000U) || *it_in > 0x0010FFFFU)
    {
      if (bufsiz < (bytes += 3U) && buf != NULL)
        return 0U;

      if (buf != NULL)
      {
        *it_out++ = REPL_8_1;
        *it_out++ = REPL_8_2;
        *it_out++ = REPL_8_3;
      }

      ++it_in;
    }
    else if (*it_in < 0x00010000U)
    {
      if (bufsiz < (bytes += 3U) && buf != NULL)
        return 0U;

      if (buf != NULL)
      {
        *it_out++ = (*it_in >> 12U & 0x0FU) | 0xE0U;
        *it_out++ = (*it_in >> 6U & 0x3FU) | 0x80U;
        *it_out++ = (*it_in & 0x3FU) | 0x80U;
      }

      ++it_in;
    }
    else
    {
      if (bufsiz < (bytes += 4U) && buf != NULL)
        return 0U;

      if (buf != NULL)
      {
        *it_out++ = (*it_in >> 18U & 0x07U) | 0xF0U;
        *it_out++ = (*it_in >> 12U & 0x3FU) | 0x80U;
        *it_out++ = (*it_in >> 6U & 0x3FU) | 0x80U;
        *it_out++ = (*it_in & 0x3FU) | 0x80U;
      }

      ++it_in;
    }
  }

  *it_out = 0U;
  return bytes;
}

size_t Utf32To16(const uint32_t *const str, uint16_t *const buf, const size_t bufsiz)
{
  size_t bytes = 2U;
  const uint32_t *it_in = str;
  uint16_t *it_out = buf;

  while (*it_in != 0U)
  {
    if (*it_in < 0x0000D800U || (*it_in > 0x0000DFFFU && *it_in < 0x00010000U))
    {
      if (bufsiz < (bytes += 2U) && buf != NULL)
        return 0U;

      if (buf != NULL)
        *it_out++ = *it_in;

      ++it_in;
    }
    else if ((*it_in > 0x0000D7FFU && *it_in < 0x0000E000U) || *it_in > 0x0010FFFFU)
    {
      if (bufsiz < (bytes += 2U) && buf != NULL)
        return 0U;

      if (buf != NULL)
        *it_out++ = REPL_CPT;

      ++it_in;
    }
    else
    {
      if (bufsiz < (bytes += 4U) && buf != NULL)
        return 0U;

      if (buf != NULL)
      {
        *it_out++ = 0x0000D800U + (((*it_in - 0x00010000U)  >> 10U) & 0x000003FFU);
        *it_out++ = 0x0000DC00U + ((*it_in - 0x00010000U) & 0x000003FFU);
      }

      ++it_in;
    }
  }

  *it_out = 0U;
  return bytes;
}

char *PtrOfPos8(const char *const str, const size_t pos)
{
  const uint8_t *it = (const uint8_t *const)str;
  size_t idx = 0U;

  for (; *it != 0U && pos > idx; ++it, ++idx)
  {
    if (*it > 0x7FU)
    {
      if (*it >= 0xC2U && *it <= 0xDFU &&
          it[1] >= 0x80U && it[1] <= 0xBFU)
        ++it;
      else if (((
                  *it == 0xE0U &&
                  it[1] >= 0xA0U && it[1] <= 0xBFU) ||
                (
                  *it >= 0xE1U && *it <= 0xECU &&
                  it[1] >= 0x80U && it[1] <= 0xBFU) ||
                (
                  *it == 0xEDU &&
                  it[1] >= 0x80U && it[1] <= 0x9FU) ||
                (
                  *it >= 0xEEU && *it <= 0xEFU &&
                  it[1] >= 0x80U && it[1] <= 0xBFU)) &&
               it[2] >= 0x80U && it[2] <= 0xBFU)
        it += 2;
      else if (((
                  *it == 0xF0U &&
                  it[1] >= 0x90U && it[1] <= 0xBFU) ||
                (
                  *it >= 0xF1U && *it <= 0xF3U &&
                  it[1] >= 0x80U && it[1] <= 0xBFU) ||
                (
                  *it == 0xF4U &&
                  it[1] >= 0x80U && it[1] <= 0x8FU)) &&
               it[2] >= 0x80U && it[2] <= 0xBFU &&
               it[3] >= 0x80U && it[3] <= 0xBFU)
        it += 3;
      else
        return NULL;
    }
  }

  if (pos == idx)
    return (char *)it;

  return NULL;
}

uint16_t *PtrOfPos16(const uint16_t *const str, const size_t pos)
{
  const uint16_t *it = str;
  size_t idx = 0U;

  for (; *it != 0U && pos > idx; ++it, ++idx)
  {
    if (*it > 0xD7FFU && *it < 0xDC00U)
    {
      ++it;
      if (*it < 0xDC00U || *it > 0xDFFFU)
        return NULL;
    }
  }

  if (pos == idx)
    return (uint16_t *)it;

  return NULL;
}

uint32_t *PtrOfPos32(const uint32_t *const str, const size_t pos)
{
  const uint32_t *it = str;
  size_t idx = 0U;

  for (; *it != 0U && pos > idx; ++it, ++idx);

  if (pos == idx)
    return (uint32_t *)it;

  return NULL;
}

int ConvertCodePoint(void *const destination, const int destination_charset, const void **const psource, const int source_charset)
{
  if (destination_charset != 8 && destination_charset != 16 && destination_charset != 32)
    return 0;

  uint8_t  *it_8  = NULL;
  uint16_t *it_16 = NULL;
  uint32_t *it_32 = NULL;
  uint32_t code_point = 0U;

  switch (source_charset)
  {
    case 8:
      it_8 = (uint8_t *const) * psource;
      if (*it_8 < 0x80U)
      {
        code_point = *it_8;
        *psource = (void *)((char *)*psource + 1);
      }
      else if (*it_8 >= 0xC2U && *it_8 <= 0xDFU)
      {
        if (it_8[1] >= 0x80U && it_8[1] <= 0xBFU)
          code_point = ((*it_8 ^ 0x000000C0U) << 6U) | (it_8[1] ^ 0x00000080U);
        else
          code_point = REPL_CPT;

        ++it_8;
        if (*it_8 != 0U)
          ++it_8;

        *psource = (void *)it_8;
      }
      else if (*it_8 >= 0xE0U && *it_8 <= 0xEFU)
      {
        if (((
               *it_8 == 0xE0U &&
               it_8[1] >= 0xA0U && it_8[1] <= 0xBFU) ||
             (
               *it_8 >= 0xE1U && *it_8 <= 0xECU &&
               it_8[1] >= 0x80U && it_8[1] <= 0xBFU) ||
             (
               *it_8 == 0xEDU &&
               it_8[1] >= 0x80U && it_8[1] <= 0x9FU) ||
             (
               *it_8 >= 0xEEU && *it_8 <= 0xEFU &&
               it_8[1] >= 0x80U && it_8[1] <= 0xBFU)) &&
            it_8[2] >= 0x80U && it_8[2] <= 0xBFU)
          code_point = ((*it_8 ^ 0x000000E0U) << 12U) | ((it_8[1] ^ 0x00000080U) << 6U) | (it_8[2] ^ 0x00000080U);
        else
          code_point = REPL_CPT;

        ++it_8;
        for (unsigned i = 0U; i < 2U && *it_8 != 0U; ++it_8, ++i);
        *psource = (void *)it_8;
      }
      else if (*it_8 >= 0xF0U && *it_8 <= 0xF4U)
      {
        if (((
               *it_8 == 0xF0U &&
               it_8[1] >= 0x90U && it_8[1] <= 0xBFU) ||
             (
               *it_8 >= 0xF1U && *it_8 <= 0xF3U &&
               it_8[1] >= 0x80U && it_8[1] <= 0xBFU) ||
             (
               *it_8 == 0xF4U &&
               it_8[1] >= 0x80U && it_8[1] <= 0x8FU)) &&
            it_8[2] >= 0x80U && it_8[2] <= 0xBFU &&
            it_8[3] >= 0x80U && it_8[3] <= 0xBFU)
          code_point = ((*it_8 ^ 0x000000F0U) << 18U) | ((it_8[1] ^ 0x00000080U) << 12U) | ((it_8[2] ^ 0x00000080U) << 6U) | (it_8[3] ^ 0x00000080U);
        else
          code_point = REPL_CPT;

        ++it_8;
        for (unsigned i = 0U; i < 3U && *it_8 != 0U; ++it_8, ++i);
        *psource = (void *)it_8;
      }
      else
      {
        code_point = REPL_CPT;
        *psource = (void *)((char *)*psource + 1);
      }
      break;

    case 16:
      it_16 = (uint16_t *const) * psource;
      if (*it_16 > 0xD7FFU && *it_16 < 0xDC00U)
      {
        const uint32_t high = *it_16++;
        if (*it_16 > 0xDBFFU && *it_16 < 0xE000U)
          code_point = (high << 10U) + *it_16 - 0x035FDC00U;
        else
          code_point = REPL_CPT;
      }
      else
        code_point = *it_16;

      if (*it_16 != 0U)
        ++it_16;

      *psource = (void *)it_16;
      break;

    case 32:
      it_32 = (uint32_t *const) * psource;
      if ((*it_32 > 0x0000D7FFU && *it_32 < 0x0000E000U) || *it_32 > 0x0010FFFFU)
        code_point = REPL_CPT;
      else
        code_point = *it_32;

      *psource = (void *)(++it_32);
      break;

    default:
      return 0;
  }

  switch (destination_charset)
  {
    case 8:
      it_8 = (uint8_t *)destination;
      if (code_point < 0x00000080U)
      {
        *it_8 = code_point;
        return 1;
      }
      else if (code_point < 0x00000800U)
      {
        *it_8++ = (code_point >> 6U & 0x1FU) | 0xC0U;
        *it_8 = (code_point & 0x3FU) | 0x80U;
        return 2;
      }
      else if (code_point < 0x00010000U)
      {
        *it_8++ = (code_point >> 12U & 0x0FU) | 0xE0U;
        *it_8++ = (code_point >> 6U & 0x3FU) | 0x80U;
        *it_8 = (code_point & 0x3FU) | 0x80U;
        return 3;
      }
      else
      {
        *it_8++ = (code_point >> 18U & 0x07U) | 0xF0U;
        *it_8++ = (code_point >> 12U & 0x3FU) | 0x80U;
        *it_8++ = (code_point >> 6U & 0x3FU) | 0x80U;
        *it_8 = (code_point & 0x3FU) | 0x80U;
        return 4;
      }

    case 16:
      it_16 = (uint16_t *)destination;
      if (code_point < 0x0000D800U || (code_point > 0x0000DFFFU && code_point < 0x00010000U))
      {
        *it_16 = code_point;
        return 1;
      }
      else
      {
        *it_16++ = 0x0000D800U + (((code_point - 0x00010000U)  >> 10U) & 0x000003FFU);
        *it_16 = 0x0000DC00U + ((code_point - 0x00010000U) & 0x000003FFU);
        return 2;
      }

    case 32:
      *(uint32_t *)destination = code_point;
      return 1;
  }

  return 0;
}

int StrCmp16(const uint16_t *str1, const uint16_t *str2)
{
  for ( ; *str1 != 0U && *str1 == *str2; ++str1, ++str2);

  return *str1 == *str2 ? 0 : (*str1 < *str2 ? -1 : 1);
}

int StrCmp32(const uint32_t *str1, const uint32_t *str2)
{
  for ( ; *str1 != 0U && *str1 == *str2; ++str1, ++str2);

  return *str1 == *str2 ? 0 : (*str1 < *str2 ? -1 : 1);
}

uint16_t *StrCpy16(uint16_t *destination, const uint16_t *source)
{
  uint16_t *it = destination;
  for ( ; *source != 0U; *it++ = *source++);
  *it = 0U;

  return destination;
}

uint32_t *StrCpy32(uint32_t *destination, const uint32_t *source)
{
  uint32_t *it = destination;
  for ( ; *source != 0U; *it++ = *source++);
  *it = 0U;

  return destination;
}

uint16_t *StrCat16(uint16_t *destination, const uint16_t *source)
{
  uint16_t *it = destination;
  for ( ; *it != 0U; ++it);
  for ( ; *source != 0U; *it++ = *source++);
  *it = 0U;

  return destination;
}

uint32_t *StrCat32(uint32_t *destination, const uint32_t *source)
{
  uint32_t *it = destination;
  for ( ; *it != 0U; ++it);
  for ( ; *source != 0U; *it++ = *source++);
  *it = 0U;

  return destination;
}
