/** \file ustring.h */

/*

Copyright (c) 2018-2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef USTRING_H_INCLUDED__
#define USTRING_H_INCLUDED__


#ifdef _MSC_VER
# define _CRT_SECURE_NO_WARNINGS // Microsoft Visual Studio supposes standard functions such as strcmp, strcpy, and strcat to be unsafe.
#endif

#include <stddef.h>   // size_t
#include <stdint.h>   // uint8_t, uint16_t, uit32_t
#include <string.h>   // strcmp, strcpy, strcat
#include <limits.h>   // CHAR_BIT


#if (CHAR_BIT) != 8 || !defined (UINT8_MAX) || !defined (UINT16_MAX) || !defined (UINT32_MAX)
# error This function library is restricted to implementations where the width of type char is 8 bits and where unsigned integer types with exact widths of 8, 16, and 32 bits exist.
#endif

#ifdef __cplusplus
extern "C" {
#endif


// Byte Order Mark
/** \brief UTF-8 BOM, first code unit */
#define BOM_8_1  '\xEF'

/** \brief UTF-8 BOM, second code unit */
#define BOM_8_2  '\xBB'

/** \brief UTF-8 BOM, third code unit */
#define BOM_8_3  '\xBF'

/** \brief code point for UTF-16 and UTF-32 BOM */
#define BOM_CPT  0xFEFFU



/** \def   B2U16(b)
  * \brief Calculate the size of a UTF-16 string in bytes to the number of code units it contains.
  *        Note that the number of code units can be greater than the number of code points in UTF-16. */
#define B2U16(b) ((size_t)(b) / 2U)

/** \def   B2U32(b)
  * \brief Calculate the size of a UTF-32 string in bytes to the number of code units it contains.
  *        Note that the number of code units equals the number of code points in UTF-32. */
#define B2U32(b) ((size_t)(b) / 4U)

/** \def   U2B16(u)
  * \brief Calculate the number of code units in a UTF-16 string to the number of bytes it contains.
  *        Note that the number of code units can be greater than the number of code points in UTF-16. */
#define U2B16(u) ((size_t)(u) * 2U)

/** \def   U2B32(u)
  * \brief Calculate the number of code units in a UTF-32 string to the number of bytes it contains.
  *        Note that the number of code units equals the number of code points in UTF-32. */
#define U2B32(u) ((size_t)(u) * 4U)



/** \brief  Change endianness of a UTF-16 string.
 *  \param  str  Pointer to the first character of a null-terminated UTF-16 string.
 *  \return str. */
uint16_t *ReverseByteOrder16(uint16_t *const str);

/** \brief  Change endianness of a UTF-32 string.
 *  \param  str  Pointer to the first character of a null-terminated UTF-32 string.
 *  \return str. */
uint32_t *ReverseByteOrder32(uint32_t *const str);



/** \brief  Determine if a UTF-8 string has a Byte Order Mark.
 *  \param  str  Pointer to the first character of a null-terminated UTF-8 string.
 *  \return 1 if the string has a BOM.
 *          0 if the string has no BOM. */
int HasBOM8(const char *const str);

/** \brief  Determine if a UTF-16 string has a Byte Order Mark.
 *  \param  str  Pointer to the first character of a null-terminated UTF-16 string.
 *  \return  1 if the string has a BOM that indicates the same endianness as the implementation.
 *          -1 if the string has a BOM that indicates a different endianness as the implementation.
 *           0 if the string has no BOM. */
int HasBOM16(const uint16_t *const str);

/** \brief  Determine if a UTF-32 string has a Byte Order Mark.
 *  \param  str  Pointer to the first character of a null-terminated UTF-32 string.
 *  \return  1 if the string has a BOM that indicates the same endianness as the implementation.
 *          -1 if the string has a BOM that indicates a different endianness as the implementation.
 *           0 if the string has no BOM. */
int HasBOM32(const uint32_t *const str);



/** \brief  Validate and count the number of Unicode code points and bytes in a UTF-8 string.
 *  \param  str    Pointer to the first character of a null-terminated UTF-8 string.
 *  \param  p_len  Pointer to a size_t receiving the number of code points in the string without terminating null.
 *  \return The number of bytes read including the terminating null. Zero value if the function failed. */
size_t Validate8(const char *const str, size_t *const p_len);

/** \brief  Validate and count the number of Unicode code points and bytes in a UTF-16 string.
 *  \param  str    Pointer to the first character of a null-terminated UTF-16 string.
 *  \param  p_len  Pointer to a size_t receiving the number of code points in the string without terminating null.
 *  \return The number of bytes read including the terminating null. Zero value if the function failed. */
size_t Validate16(const uint16_t *const str, size_t *const p_len);

/** \brief  Validate and count the number of Unicode code points and bytes in a UTF-32 string.
 *  \param  str    Pointer to the first character of a null-terminated UTF-32 string.
 *  \param  p_len  Pointer to a size_t receiving the number of code points in the string without terminating null.
 *  \return The number of bytes read including the terminating null. Zero value if the function failed. */
size_t Validate32(const uint32_t *const str, size_t *const p_len);



/** \brief  Convert a CP850 string to a UTF-32 string.
 *  \param  str     Pointer to the first character of a null-terminated CP850 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the UTF-32 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Cp850ToUtf32(const char *const str, uint32_t *const buf, const size_t bufsiz);

/** \brief  Convert a Windows-1252 string to a UTF-32 string.
 *  \param  str     Pointer to the first character of a null-terminated Windows-1252 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the UTF-32 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Cp1252ToUtf32(const char *const str, uint32_t *const buf, const size_t bufsiz);

/** \brief  Convert a UTF-8 string to a UTF-32 string.
 *  \param  str     Pointer to the first character of a null-terminated UTF-8 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the UTF-32 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Utf8To32(const char *const str, uint32_t *const buf, const size_t bufsiz);

/** \brief  Convert a UTF-16 string to a UTF-32 string.
 *  \param  str     Pointer to the first character of a null-terminated UTF-16 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the UTF-32 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Utf16To32(const uint16_t *const str, uint32_t *const buf, const size_t bufsiz);

/** \brief  Convert a UTF-32 string to a CP850 string.
 *  \param  str     Pointer to the first character of a null-terminated UTF-32 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the CP850 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Utf32ToCP850(const uint32_t *const str, char *const buf, const size_t bufsiz);

/** \brief  Convert a UTF-32 string to a Windows-1252 string.
 *  \param  str     Pointer to the first character of a null-terminated UTF-32 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the Windows-1252 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Utf32ToCP1252(const uint32_t *const str, char *const buf, const size_t bufsiz);

/** \brief  Convert a UTF-32 string to a UTF-8 string.
 *  \param  str     Pointer to the first character of a null-terminated UTF-32 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the UTF-8 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Utf32To8(const uint32_t *const str, char *const buf, const size_t bufsiz);

/** \brief  Convert a UTF-32 string to a UTF-16 string.
 *  \param  str     Pointer to the first character of a null-terminated UTF-32 string.
 *  \param  buf     Pointer to the first element of a buffer array receiving the UTF-16 string.
 *  \param  bufsiz  Size of the buffer in bytes.
 *  \return The number of bytes written including the terminating null. Zero value if the function failed.
 *          If buf is NULL the function returns the required size for the buffer in bytes amd the bufsiz parameter will be ignored. */
size_t Utf32To16(const uint32_t *const str, uint16_t *const buf, const size_t bufsiz);



/** \brief  Get the pointer to a Unicode code point in a UTF-8 string.
 *  \param  str  Pointer to the first character of a null-terminated UTF-8 string.
 *  \param  pos  Zero-based position of the code point.
 *  \return Pointer to the first code unit of the code point in the string, specified by its position.
 *          NULL pointer if the function failed. */
char *PtrOfPos8(const char *const str, const size_t pos);

/** \brief  Get the pointer to a Unicode code point in a UTF-16 string.
 *  \param  str  Pointer to the first character of a null-terminated UTF-16 string.
 *  \param  pos  Zero-based position of the code point.
 *  \return Pointer to the first code unit of the code point in the string, specified by its position.
 *          NULL pointer if the function failed. */
uint16_t *PtrOfPos16(const uint16_t *const str, const size_t pos);

/** \brief  Get the pointer to a Unicode code point in a UTF-32 string.
 *  \param  str  Pointer to the first character of a null-terminated UTF-32 string.
 *  \param  pos  Zero-based position of the code point.
 *  \return Pointer to the code point in the string, specified by its position.
 *          NULL pointer if the function failed. */
uint32_t *PtrOfPos32(const uint32_t *const str, const size_t pos);



/** \brief  Convert a code point from one unicode charset to another.
 *  \param  destination          Pointer to an element of a buffer array receiving the converted code point.
 *  \param  destination_charset  Identifier of the charset that the codepoint pointed to by *psource shall be converted.
 *                               (8 for UTF-8, 16 for UTF-16, 32 for UTF-32)
 *  \param  psource              Address of a pointer to the first code unit of the code point to be converted.
 *                               If the function succeeded the pointer value of *psource points to the first code unit of the next code point.
 *  \param  source_charset       Identifier of the charset of the codepoint pointed to by *psource.
 *                               (8 for UTF-8, 16 for UTF-16, 32 for UTF-32)
 *  \return Number of code units written to destination. Zero value if the function failed. */
int ConvertCodePoint(void *const destination, const int destination_charset, const void **const psource, const int source_charset);



/** \brief  Compares UTF-8 string str1 to UTF-8 string str2.
 *  \param  str1  Pointer to the first character of a null-terminated UTF-8 string.
 *  \param  str1  Pointer to the first character of a null-terminated UTF-8 string.
 *  \return <0 if the first character that does not match has a lower value in str1 than in str2.
 *           0 if the contents of both strings are equal.
 *          >0 if the first character that does not match has a higher value in str1 than in str2. */
inline int StrCmp8(const char *str1, const char *str2)
{
  return strcmp(str1, str2);
}

/** \brief  Compares UTF-16 string str1 to UTF-16 string str2.
 *  \param  str1  Pointer to the first character of a null-terminated UTF-16 string.
 *  \param  str1  Pointer to the first character of a null-terminated UTF-16 string.
 *  \return <0 if the first character that does not match has a lower value in str1 than in str2.
 *           0 if the contents of both strings are equal.
 *          >0 if the first character that does not match has a higher value in str1 than in str2. */
int StrCmp16(const uint16_t *str1, const uint16_t *str2);

/** \brief  Compares UTF-32 string str1 to UTF-32 string str2.
 *  \param  str1  Pointer to the first character of a null-terminated UTF-32 string.
 *  \param  str1  Pointer to the first character of a null-terminated UTF-32 string.
 *  \return <0 if the first character that does not match has a lower value in str1 than in str2.
 *           0 if the contents of both strings are equal.
 *          >0 if the first character that does not match has a higher value in str1 than in str2. */
int StrCmp32(const uint32_t *str1, const uint32_t *str2);



/** \brief  Copies the UTF-8 string pointed by source into the UTF-8 array pointed by destination, including the terminating null.
 *  \param  destination  Pointer to the first character of the UTF-8 destination array where the content is to be copied.
 *  \param  source       Pointer to the first character of the null-terminated UTF-8 string to be copied.
 *  \return destination. */
inline char *StrCpy8(char *destination, const char *source)
{
  return strcpy(destination, source);
}

/** \brief  Copies the UTF-16 string pointed by source into the UTF-16 array pointed by destination, including the terminating null.
 *  \param  destination  Pointer to the first character of the UTF-16 destination array where the content is to be copied.
 *  \param  source       Pointer to the first character of the null-terminated UTF-16 string to be copied.
 *  \return destination. */
uint16_t *StrCpy16(uint16_t *destination, const uint16_t *source);

/** \brief  Copies the UTF-32 string pointed by source into the UTF-32 array pointed by destination, including the terminating null.
 *  \param  destination  Pointer to the first character of the UTF-32 destination array where the content is to be copied.
 *  \param  source       Pointer to the first character of the null-terminated UTF-32 string to be copied.
 *  \return destination. */
uint32_t *StrCpy32(uint32_t *destination, const uint32_t *source);



/** \brief  Appends the UTF-8 string pointed by source to the UTF-8 array pointed by destination, including the terminating null,
 *          where destination and source shall not overlap.
 *  \param  destination  Pointer to the first character of the UTF-8 destination array where the content is to be appended.
 *  \param  source       Pointer to the first character of the null-terminated UTF-8 string to be appended.
 *  \return destination. */
inline char *StrCat8(char *destination, const char *source)
{
  return strcat(destination, source);
}

/** \brief  Appends the UTF-16 string pointed by source to the UTF-16 array pointed by destination, including the terminating null,
 *          where destination and source shall not overlap.
 *  \param  destination  Pointer to the first character of the UTF-16 destination array where the content is to be appended.
 *  \param  source       Pointer to the first character of the null-terminated UTF-16 string to be appended.
 *  \return destination. */
uint16_t *StrCat16(uint16_t *destination, const uint16_t *source);

/** \brief  Appends the UTF-32 string pointed by source to the UTF-32 array pointed by destination, including the terminating null,
 *          where destination and source shall not overlap.
 *  \param  destination  Pointer to the first character of the UTF-32 destination array where the content is to be appended.
 *  \param  source       Pointer to the first character of the null-terminated UTF-32 string to be appended.
 *  \return destination. */
uint32_t *StrCat32(uint32_t *destination, const uint32_t *source);



// pointer types to be used for type casts
/** \typedef const void **const
  * \brief   Type of parameter source of the ConvertCodePoint function. */
typedef  const void **const  CCP_SRC_T;

/** \typedef void *const
  * \brief   Type of parameter destination of the ConvertCodePoint function. */
typedef        void  *const  CCP_DST_T;


#ifdef __cplusplus
}
#endif

#endif // USTRING_H_INCLUDED__
